/*
 *  InputManager.java
 *  Created on January 29, 2002, 1:31 PM
 *  Version 1beta0
 *  @author  Marco Parvis
 */

/*
    Questo e' InputManager una parte di Instrument Training Project ITP
    Copyright (C) <Politecnico di Torino 2002> <Marco Parvis>
 
    This is InputManager part of the Instrument Training Project ITP
    Copyright (C) <Politecnico di Torino 2002> <Marco Parvis>
 
 */

/*
    Questo  programma e'  software  libero; e'  lecito redistribuirlo  o
    modificarlo secondo i termini  della Licenza Pubblica Generica GNU
    come e' pubblicata dalla Free  Software Foundation; o la versione 2
    della licenza o (a propria scelta) una versione successiva.
 
    Questo programma  e' distribuito nella  speranza che sia  utile, ma
    SENZA  ALCUNA GARANZIA;  senza  neppure la  garanzia implicita  di
    NEGOZIABILITA'  o di  APPLICABILITA' PER  UN PARTICOLARE  SCOPO.  Si
    veda la Licenza Pubblica Generica GNU per avere maggiori dettagli.
 
    Questo  programma deve  essere  distribuito assieme  ad una  copia
    della Licenza Pubblica Generica GNU;  in caso contrario, se ne puo'
    ottenere  una scrivendo  alla Free  Software Foundation,  Inc., 59
    Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
    L'autore puo' essere contattato a parvis@polito.it
    Dipartimento di Elettronica Politecnico di Torino
    Corso Duca degli Abruzzi, 24, 10129 Torino Italy
 */

/*
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
 
    The author can be contacted at parvis@polito.it
    Dipartimento di Elettronica Politecnico di Torino
    Corso Duca degli Abruzzi, 24, 10129 Torino Italy
 
 */


/*
 * InpuManager.java
 * This class manages the inputADC and notifies observers with a short[] containing
 * the samples
 */
package JavaScope;

import java.util.*;
import javax.sound.sampled.*;
import java.nio.*;




public class InputManager extends Observable {
    private boolean bigEndianPlatform;
    private InputBufferDriver bd;
    final int bufferLenInBytes=35280; //input buffer size 70560=44100*2*2*0.2 0.2s
    final int bufferLenInWords=bufferLenInBytes/2;
    
    //implement a double buffer of read data
    short[] bufferA=new short[bufferLenInWords];
    short[] bufferB=new short[bufferLenInWords];
    int bufferToBeUpdated=0;
    
    
    private int offsetCh1=0,offsetCh2=0;
    private boolean doOffsetRemoval=false;
    /** Creates new InputManager */
    public InputManager() {
        ByteOrder bo=ByteOrder.nativeOrder();
        if (bo==ByteOrder.BIG_ENDIAN)
            bigEndianPlatform=true;
        else
            bigEndianPlatform=false;
        
        this.bd=new InputBufferDriver();
        
    }
    public void stopCapture(){
        
        bd.stopCapture();
    }
    
    public void useInputOffset(boolean use){
        doOffsetRemoval=use;
    };
    public void setInputOffset(short newOffsetCh1,short newOffsetCh2){
        offsetCh1=(int)newOffsetCh1;
        offsetCh2=(int)newOffsetCh2;
    };
    
    
    
    
    public short[] getLastBuffer(){
        if (bufferToBeUpdated==0)
            return bufferB;
        else
            return bufferA;
    }
    public void notifyObservers(Object o){
        setChanged();
        super.notifyObservers(o);
    }
    
    public void notifyObservers(){
        setChanged();
        super.notifyObservers();
    }
    
    class InputBufferDriver  extends Thread{
        private long numberOfChunkSoFar=0;
        private int bufferToBeUpdated=0;
        final int bufferLenInBytes=35280; //input buffer size 70560=44100*2*2*0.2 0.2s
        final int bufferLenInWords=bufferLenInBytes/2;
        private AudioInputStream audioInputStream;
        private TargetDataLine line;
        
        private boolean doCapture;
        
        public InputBufferDriver(){
            this.doCapture=true;
            audioInputStream = null;
            // define the required attributes for our line,
            // and make sure a compatible line is supported.
            AudioFormat format = getFormat();
            DataLine.Info info = new DataLine.Info(TargetDataLine.class,
            format);
            
            if (!AudioSystem.isLineSupported(info)) {
                System.out.println("Line matching " + info + " not supported.");
                return;
            }
            // get and open the target data line for capture.
            try {
                line = (TargetDataLine) AudioSystem.getLine(info);
                line.open(format, bufferLenInBytes*2); //line.getBufferSize());
                
            } catch (LineUnavailableException ex) {
                System.out.println("Unable to open the line: " + ex);
                return;
            } catch (SecurityException ex) {
                System.out.println(ex.toString());
                return;
            } catch (Exception ex) {
                System.out.println(ex.toString());
                return;
            }
/* this not yet supported
            try{
                FloatControl gainControl =
                  (FloatControl) line.getControl(FloatControl.Type.VOLUME);
                double value=0.2;
                float dB = (float)
                  (Math.log(value==0.0?0.0001:value)/Math.log(10.0)*20.0);
                gainControl.setValue(dB);
                 //
 
            } catch (Exception ex) {
                System.out.println(ex.toString());
 
            }
 
 */
            
            
            
            this.start(); //start the capture thread
        }//of craator
        public void stopCapture(){
            doCapture=false;
        }
        public void run(){
            int numBytesRead;
            //allocate space for the chunks
            byte[] data = new byte[bufferLenInBytes];
            //we use a direct buffer to speedup the processing (maybe...)
            ByteBuffer bb = ByteBuffer.allocateDirect(bufferLenInBytes);
            //however we must use the buffer with the correct order
            if (!bigEndianPlatform)
                bb.order(ByteOrder.LITTLE_ENDIAN);
            //this view allow an easy data retrieval
            ShortBuffer sb=bb.asShortBuffer();
            
            line.start();
            while (doCapture) {
                if((numBytesRead = line.read(data, 0, bufferLenInBytes)) == -1) {
                    break;
                }
                numberOfChunkSoFar+=1;
                bb.clear();
                sb.clear();
                bb.put(data);
                
                
                
                
                if (bufferToBeUpdated==0){
                    sb.get(bufferA);
                    bufferToBeUpdated=1;
                    if (doOffsetRemoval){
                        for (int i=0;i<bufferA.length/2;i++){
                            int x=bufferA[2*i]-offsetCh1;
                            if (x>Short.MAX_VALUE)
                                bufferA[2*i]=Short.MAX_VALUE;
                            else if (x<Short.MIN_VALUE)
                                bufferA[2*i]=Short.MIN_VALUE;
                            else
                                bufferA[2*i]=(short)x;
                            x=bufferA[2*i+1]-offsetCh2;
                            if (x>Short.MAX_VALUE)
                                bufferA[2*i+1]=Short.MAX_VALUE;
                            else if (x<Short.MIN_VALUE)
                                bufferA[2*i+1]=Short.MIN_VALUE;
                            else
                                bufferA[2*i+1]=(short)x;
                        }
                        
                    }
                    
                    
                    
                    
                    notifyObservers((Object)bufferA);
                }
                else{
                    sb.get(bufferB);
                    bufferToBeUpdated=0;
                    if (doOffsetRemoval){
                        for (int i=0;i<bufferB.length/2;i++){
                            int x=bufferB[2*i]-offsetCh1;
                            if (x>Short.MAX_VALUE)
                                bufferB[2*i]=Short.MAX_VALUE;
                            else if (x<Short.MIN_VALUE)
                                bufferB[2*i]=Short.MIN_VALUE;
                            else
                                bufferB[2*i]=(short)x;
                            x=bufferB[2*i+1]-offsetCh2;
                            if (x>Short.MAX_VALUE)
                                bufferB[2*i+1]=Short.MAX_VALUE;
                            else if (x<Short.MIN_VALUE)
                                bufferB[2*i+1]=Short.MIN_VALUE;
                            else
                                bufferB[2*i+1]=(short)x;
                            
                        }
                    }
                    
                    notifyObservers((Object)bufferB);
                }
                
            }
            // we reached the end of the stream.  stop and close the line.
            line.stop();
            line.close();
            line = null;
            
        }
        
        private AudioFormat getFormat() {
            AudioFormat.Encoding encoding = AudioFormat.Encoding.PCM_SIGNED;
            int channels=2;
            int sampleSize=16;
            float rate=44100;
            
            boolean bigEndian=bigEndianPlatform; //true;
            return new AudioFormat(encoding,
            rate,
            sampleSize,
            channels,
            (sampleSize/8)*channels,
            rate,
            bigEndian);
        }
    }//of inputDriver
    
    
    
    
    
}//of inputManager
