/*
 *  OutputManager.java
 *  Created on January 29, 2002, 1:31 PM
 *  Version 1beta0
 *  @author  Marco Parvis
 */

/*
    Questo e' OutputManager una parte di Instrument Training Project ITP
    Copyright (C) <Politecnico di Torino 2002> <Marco Parvis>
 
    This is OutputManager part of the Instrument Training Project ITP
    Copyright (C) <Politecnico di Torino 2002> <Marco Parvis>
 
 */

/*
    Questo  programma e'  software  libero; e'  lecito redistribuirlo  o
    modificarlo secondo i termini  della Licenza Pubblica Generica GNU
    come e' pubblicata dalla Free  Software Foundation; o la versione 2
    della licenza o (a propria scelta) una versione successiva.
 
    Questo programma  e' distribuito nella  speranza che sia  utile, ma
    SENZA  ALCUNA GARANZIA;  senza  neppure la  garanzia implicita  di
    NEGOZIABILITA'  o di  APPLICABILITA' PER  UN PARTICOLARE  SCOPO.  Si
    veda la Licenza Pubblica Generica GNU per avere maggiori dettagli.
 
    Questo  programma deve  essere  distribuito assieme  ad una  copia
    della Licenza Pubblica Generica GNU;  in caso contrario, se ne puo'
    ottenere  una scrivendo  alla Free  Software Foundation,  Inc., 59
    Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
    L'autore puo' essere contattato a parvis@polito.it
    Dipartimento di Elettronica Politecnico di Torino
    Corso Duca degli Abruzzi, 24, 10129 Torino Italy
 */

/*
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
 
    The author can be contacted at parvis@polito.it
    Dipartimento di Elettronica Politecnico di Torino
    Corso Duca degli Abruzzi, 24, 10129 Torino Italy
 
 */


/*
 * This class manages the output channel of the audio board
 * It can play signals described by a sequence af samples
 * and audio clips contained in wav files
 */
package JavaScope;




//import java.lang.*;
import java.util.*;
import java.io.*;
import javax.sound.sampled.*;
import javax.sound.midi.*;

import java.nio.*;


import java.awt.*;
//import java.awt.event.*;
import javax.swing.*;
import java.text.*;
public class OutputManager extends Observable {
    
    public final int OUT_STARTED=1;
    public final int OUT_STOPPING=2;
    public final int OUT_STOPPED=3;
    public final int OUT_CHUNK_DONE=4;
    
    public final int OUT_ERROR=100;
    public final int OUT_PARAM_ERROR=101;
    public final int OUT_UNKNOWN=1000;
    
    public final int STATUS_OUT_NORMAL=1;
    public final int STATUS_OUT_STOPPING=2;
    
    
    public final Integer OBJ_OUT_STARTED=new Integer(OUT_STARTED);
    public final Integer OBJ_OUT_STOPPING=new Integer(OUT_STOPPING);
    public final Integer OBJ_OUT_STOPPED=new Integer(OUT_STOPPED);
    public final Integer OBJ_OUT_ERROR=new Integer(OUT_ERROR);
    public final Integer OBJ_OUT_CHUNK_DONE=new Integer(OUT_CHUNK_DONE);
    public final Integer OBJ_OUT_PARAM_ERROR=new Integer(OUT_PARAM_ERROR);
    public final Integer OBJ_OUT_UNKNOWN=new Integer(OUT_UNKNOWN);
    
    public int status;
    private OutputBufferDriver bd=null;
    private File currentClip;
    private byte[] prepareData=new byte[2];
    private java.util.Timer timerMonitor;
    private OutputMonitor outputMonitor;
    private boolean bigEndianPlatform;
    
    
    public OutputManager() {
        ByteOrder bo=ByteOrder.nativeOrder();
        if (bo==ByteOrder.BIG_ENDIAN)
            bigEndianPlatform=true;
        else
            bigEndianPlatform=false;
        
        status=STATUS_OUT_NORMAL;
        outputMonitor=new OutputMonitor();
        timerMonitor=new java.util.Timer();
        timerMonitor.scheduleAtFixedRate(outputMonitor,200,200);
    }
    public void processDriverMessage(int messageId){
        
        switch (messageId){
            case OUT_STARTED:
                status=STATUS_OUT_NORMAL;
                setChanged();
                this.notifyObservers(OBJ_OUT_STARTED);
                break;
            case OUT_STOPPING:
                status=STATUS_OUT_STOPPING;
                break;
            case OUT_STOPPED:
                status=STATUS_OUT_NORMAL;
                setChanged();
                this.notifyObservers(OBJ_OUT_STOPPED);
                break;
            case OUT_ERROR:
                status=STATUS_OUT_NORMAL;
                setChanged();
                this.notifyObservers(OBJ_OUT_ERROR);
                break;
            case OUT_CHUNK_DONE:
                status=STATUS_OUT_NORMAL;
                setChanged();
                this.notifyObservers(OBJ_OUT_CHUNK_DONE);
                break;
            case OUT_PARAM_ERROR:
                status=STATUS_OUT_NORMAL;
                setChanged();
                this.notifyObservers(OBJ_OUT_PARAM_ERROR);
                break;
            default:
                status=STATUS_OUT_NORMAL;
                setChanged();
                this.notifyObservers(OBJ_OUT_UNKNOWN);
                break;
        }
        
    }
    
    
    /*--------   WAVEFORM FUNCTIONS -------------------------------------*/
    
    public boolean setWaveforms(short[] ch1,short[] ch2){
        if (ch1.length!=ch2.length){
            return false;
        }
        int bufferLenInBytes=4*ch1.length;
        ByteBuffer bb = ByteBuffer.allocateDirect(bufferLenInBytes);
        //however we must use the buffer with the correct order
        if (!bigEndianPlatform)
            bb.order(ByteOrder.LITTLE_ENDIAN);
        //this view allow an easy data retrieval
        ShortBuffer sb=bb.asShortBuffer();
        bb.clear();
        sb.clear();
        for (int i=0;i<ch1.length;i++){
            sb.put(ch1[i]);
            sb.put(ch2[i]);
        }
        prepareData=new byte[bufferLenInBytes];
        bb.get(prepareData);
        
        return true;
        
    }
    
    public boolean updateWaveforms(short[] ch1,short[] ch2){
        if (bd==null)
            return false;
        
        if (ch1.length!=ch2.length){
            processDriverMessage(OUT_PARAM_ERROR);
            return false;
        }
        if (ch1.length!=bd.actualDataLen()){
            processDriverMessage(OUT_PARAM_ERROR);
            return false;
        }
        int bufferLenInBytes=4*ch1.length;
        ByteBuffer bb = ByteBuffer.allocateDirect(bufferLenInBytes);
        //however we must use the buffer with the correct order
        if (!bigEndianPlatform)
            bb.order(ByteOrder.LITTLE_ENDIAN);
        //this view allow an easy data retrieval
        ShortBuffer sb=bb.asShortBuffer();
        bb.clear();
        sb.clear();
        for (int i=0;i<ch1.length;i++){
            sb.put(ch1[i]);
            sb.put(ch2[i]);
        }
        prepareData=new byte[bufferLenInBytes];
        bb.get(prepareData);
        
        return this.bd.updateWaveforms(prepareData);
    }
    
    
    public boolean startWav(short[] ch1,short[] ch2){
        if (!setWaveforms(ch1,ch2))
            return false;
        return startWav();
    }
    public boolean startWav(){
        if (bd!=null)
            if (bd.isAlive()){
                return false;}
        
        this.bd=new OutputBufferDriver();
        return bd.startPlay(prepareData);
    }
    
    
    /**************************************************************************/
    
    /* CLIP FUNCTIONS ----------------------------------------------------*/
    
    public void setClip(String s){
        this.currentClip=new File(s);
    }
    public void setClip(File f){
        this.currentClip=f;
    }
    public boolean startClip(String s){
        this.currentClip=new File(s);
        return startClip();
    }
    public boolean startClip(File f){
        this.currentClip=f;
        return startClip();
    }
    
    public boolean startClip(){
        //do some preliminary checks
        if (currentClip==null){
            return false;}
        
        if (bd!=null)
            if (bd.isAlive()){
                return false;}
        
        this.bd=new OutputBufferDriver();
        if (bd.loadClip(currentClip)){
            bd.playClip();
            return true;}
        else{
            return false;}
    }
    
    
    /**************************************************************/
    
    public void stopPlay(){
        if (bd!=null)
            bd.stopPlay();
    }
    
    public boolean isStopped(){
        if (bd!=null){
            return !bd.isAlive();}
        else{
            return true;}
    }
    
    
    public static void main(String args[]) {
        new OutputManager();
    }
    
    
    
    //inner monitor task **********************************************
    class OutputMonitor extends TimerTask{
        public OutputMonitor() {
        }
        public void run() {
            if (status==STATUS_OUT_STOPPING){
                if (bd!=null){
                    if (!bd.isAlive()){
                        processDriverMessage(OUT_STOPPED);
                    }else{
                    }
                    return;
                }
            }else{
                return;
            }
        }//method run
    }//class monitor
    
    class OutputBufferDriver  extends Thread{
        private Integer notif=new Integer(0);
        private int bufferLenInBytes=35280;
        private SourceDataLine line;
        byte[] data1;
        byte[] data2;
        private boolean use2buffer=false;
        private boolean swapBuffers=false;
        private boolean doPlay;
        //clip variables
        private Sequencer sequencer;
        private Object currentSound;
        private File currentName;
        private Clip currentClip;
        private double duration;
        private boolean isClip;
        //end clip related variables
        
        public OutputBufferDriver(){
        }
        /*Open the line for play*/
        public void openSequencer() {
            try {
                sequencer = MidiSystem.getSequencer();
            /*if (sequencer instanceof Synthesizer) {
                                synthesizer = (Synthesizer)sequencer;
                                channels = synthesizer.getChannels();
                        }
             */
            } catch (Exception ex) { ex.printStackTrace(); return; }
            //sequencer.addMetaEventListener(this);
        }//end of openSequeuncer
        private boolean openLine(int bufferLen){
            AudioFormat format = getFormat(); //get format 44.1kHz, 16 bit stereo
            DataLine.Info info = new DataLine.Info(SourceDataLine.class,
            format);
            if (!AudioSystem.isLineSupported(info)) {
                System.out.println("BUFFER DRIVER: Line matching " + info + " not supported.");
                processDriverMessage(OUT_ERROR);
                return false;
            }
            // get and open the source data line for playback.
            try {
                line = (SourceDataLine) AudioSystem.getLine(info);
                line.open(format, bufferLenInBytes);
                /*
                FloatControl gainControl =
                  (FloatControl) line.getControl(FloatControl.Type.MASTER_GAIN);
                double value=0.2;
                float dB = (float)
                  (Math.log(value==0.0?0.0001:value)/Math.log(10.0)*20.0);
                gainControl.setValue(dB);
                 */
            }
            catch (LineUnavailableException ex) {
                System.out.println("BUFFER DRIVER: Unable to open the line: " + ex);
                processDriverMessage(OUT_ERROR);
                return false;
            }
            return true;
        }//end of open line
        
        public boolean startPlay(byte[] newData){
            //stop the thread and close the line if required
            this.data1=newData;
            this.data2=newData;
            this.bufferLenInBytes=data1.length;
            //start the thread
            this.doPlay=true;
            this.isClip=false;
            if (this.openLine(bufferLenInBytes)){
                String aia="WAV_"+System.currentTimeMillis();
                this.setName(aia);
                this.start(); //start the play thread
                return true;
            }
            else{
                System.out.println("BUFFER DRIVER: Could not open the line !!");
                processDriverMessage(OUT_ERROR);
                return false;
            }
        }//end of startPlay
        
        public void stopPlay(){
            doPlay=false;
        }//of stopPlay
        /*Update the waveforms on the fly provided that the buffer lenght
         *does not need to be changed
         */
        public boolean updateWaveforms(byte[] b){
            if (b.length==data1.length){
                if (use2buffer)
                    this.data1=b;
                else
                    this.data2=b;
                swapBuffers=true;
                return true;
            }else
                return false;
        }//end of updateWaveforms
        public int actualDataLen(){
            return data1.length/4;
        }
        public void run(){
            int numBytesWrote;
            if (isClip){
                currentClip.loop(Clip.LOOP_CONTINUOUSLY );
                try { sleep(99); } catch (Exception e) { }
                while ((currentClip.isActive()) & doPlay ) {
                    try { sleep(199); }
                    catch (Exception e) {
                        break;
                    }
                }
                currentClip.stop();
                currentClip.close();
                try {
                    if (currentSound instanceof AudioInputStream) {
                        ((AudioInputStream)currentSound).close();
                    }
                    else if (currentSound instanceof BufferedInputStream) {
                        ((BufferedInputStream)currentSound).close();
                    }
                }
                catch (IOException ioe) {
                    System.out.println("BUFFER DRIVER: error closing stream");
                    
                }
                currentClip=null;
                currentSound=null;
                processDriverMessage(OUT_STOPPING);
                
            }//of isClip
            else{
                line.start();
                processDriverMessage(OUT_STARTED);
                while (doPlay){
                    try {
                        if (swapBuffers){
                            use2buffer=!use2buffer;
                            swapBuffers=false;
                        }
                        if (use2buffer)
                            numBytesWrote=line.write(data2, 0, bufferLenInBytes);
                        else
                            numBytesWrote=line.write(data1, 0, bufferLenInBytes);
                        processDriverMessage(OUT_CHUNK_DONE);
                        
                    }
                    catch (Exception e) {
                        System.out.println("BUFFER DRIVER: Error during playback: " + e);
                        processDriverMessage(OUT_ERROR);
                        break;
                    }
                }//of while(doPlay)
                line.stop();
                line.close();
                line = null;
                processDriverMessage(OUT_STOPPING);
            }//of !isClip
            return;
            
        }//of method run
        
        private AudioFormat getFormat() {
            AudioFormat.Encoding encoding = AudioFormat.Encoding.PCM_SIGNED;
            int channels=2;
            int sampleSize=16;
            float rate=44100;
            boolean bigEndian=bigEndianPlatform; //false;
            return new AudioFormat(encoding,
            rate,
            sampleSize,
            channels,
            (sampleSize/8)*channels,
            rate,
            bigEndian);
        }//of getFormat
        
        public boolean loadClip(File fileName) {
            duration = 0.0;
            currentName = fileName;
            try {
                currentSound = AudioSystem.getAudioInputStream(fileName);
            } catch(Exception e1) {
                // load midi & rmf as inputstreams for now
                try {
                    FileInputStream is = new FileInputStream(fileName);
                    currentSound = new BufferedInputStream(is, 1024);
                } catch (Exception e3) {
                    e3.printStackTrace();
                    currentSound = null;
                    return false;
                }
            }//of catch
            if (currentSound instanceof AudioInputStream) {
                try {
                    AudioInputStream stream = (AudioInputStream) currentSound;
                    AudioFormat format = stream.getFormat();
                    if (format.getEncoding() != AudioFormat.Encoding.PCM_SIGNED) {
                        JOptionPane.showMessageDialog(
                        null,
                        "This clip cannot be loaded due to a format. \n"+
                        "This program supports PCM only",
                        "Alert",
                        JOptionPane.ERROR_MESSAGE);
                        
                        currentSound = null;
                        
                        System.out.println("BUFFER DRIVER: Improper clip format NEED PCM");
                        
                        return false;
                    }
                    DataLine.Info info = new DataLine.Info(
                    Clip.class,
                    stream.getFormat(),
                    ((int) stream.getFrameLength() *
                    format.getFrameSize()));
                    
                    currentClip = (Clip) AudioSystem.getLine(info);
                    currentClip.open(stream);
                    stream.close();
                } catch (Exception ex) {
                    
                    JOptionPane.showMessageDialog(
                    null,
                    "This clip cannot be loaded due to a format or size problem. \n"+
                    "If the problem is the size, you can try playing it with a standard"+
                    " (non-looping) player\n"+
                    "The error message returned by the java layer is\n"+ex.getMessage(),
                    "Alert",
                    JOptionPane.ERROR_MESSAGE);
                    
                    //ex.printStackTrace();
                    currentSound = null;
                    return false;
                }//of catch
            }//of AudioInputStream
            else if (currentSound instanceof BufferedInputStream) {
                try {
                    sequencer.open();
                    sequencer.setSequence((BufferedInputStream) currentSound);
                } catch (InvalidMidiDataException imde) {
                    System.out.println("BUFFER DRIVER: Unsupported audio file.");
                    currentSound = null;
                    return false;
                } catch (Exception ex) {
                    ex.printStackTrace();
                    currentSound = null;
                    return false;
                }
            }//of bufferedInpurStream
            return true;
        }//of loadClip
        
        public double getDuration() {
            double duration = 0.0;
            if (currentSound instanceof Sequence) {
                duration = ((Sequence) currentSound).getMicrosecondLength() / 1000000.0;
            }  else if (currentSound instanceof BufferedInputStream) {
                duration = sequencer.getMicrosecondLength() / 1000000.0;
            } else if (currentSound instanceof Clip) {
                Clip clip = (Clip) currentSound;
                duration = clip.getBufferSize() /
                (clip.getFormat().getFrameSize() * clip.getFormat().getFrameRate());
            }
            return duration;
        }//of getDuration
        public double getSeconds() {
            double seconds = 0.0;
            if (currentSound instanceof Clip) {
                Clip clip = (Clip) currentSound;
                seconds = clip.getFramePosition() / clip.getFormat().getFrameRate();
            } else if ( (currentSound instanceof Sequence) || (currentSound instanceof BufferedInputStream) ) {
                try {
                    seconds = sequencer.getMicrosecondPosition() / 1000000.0;
                } catch (IllegalStateException e){
                    System.out.println("BUFFER DRIVER: IllegalStateException "+
                    "on sequencer.getMicrosecondPosition(): " + e);
                }
            }
            return seconds;
        }//of getSeconds
        
        public void playClip() {
            boolean midiEOM,audioEOM,bump;
            midiEOM = audioEOM = bump = false;
            Thread thread=null;
            if (currentClip instanceof Sequence || currentSound instanceof BufferedInputStream && thread != null) {
                sequencer.start();
                while (!midiEOM && thread != null && !bump) {
                    try { sleep(99); } catch (Exception e) {break;}
                }
                sequencer.stop();
                sequencer.close();
            } else if (currentClip instanceof Clip ) {
                isClip=true;
                doPlay=true;
                String debugString="Clip_"+System.currentTimeMillis();
                this.setName(debugString);
                this.start();
            }
        }//of playClip
        
    }//of outputDriver
    
}//of OutputManager
